__author__ = "ardevelop"

try:
    import ujson as json
except ImportError:
    import json

import validation

OBJECT_DELIMITER = "."


class ModelObject(object):
    def __init__(self, prefix=None):
        self.dict = {}
        self.prefix = prefix + OBJECT_DELIMITER if prefix else ""

    def __getattr__(self, field):
        try:
            return self.dict[field]
        except KeyError:
            return None

    def __getitem__(self, field):
        try:
            return self.dict[field]
        except KeyError:
            return None

    def __setitem__(self, field, value):
        self.dict[field] = value

    def __repr__(self):
        return str(self.dict)

    def describe(self):
        pass

    def resolve(self):
        pass

    def set(self, field, value):
        self.dict[field] = value

    def append(self, field, value):
        try:
            self.dict[field].append(value)
        except KeyError:
            self.dict[field] = [value]
        except AttributeError:
            raise Exception("Field already have value different from list")

    def get_or_create_object(self, field):
        try:
            return self.dict[field]
        except KeyError:
            obj = ModelObject(field)
            self.dict[field] = obj
            return obj

    def get_or_create_item(self, field, index):
        try:
            return self.dict[field][index]
        except KeyError:
            if index == 0:
                item = ModelObject(field)
                self.dict[field] = [item]
                return item
            else:
                raise Exception("Cannot start array from index different from 0")
        except IndexError:
            array = self.dict[field]
            if index == len(array):
                item = ModelObject(field)
                array.append(item)
                return item
            else:
                raise Exception("Cannot create array with this index")

    def get(self, *fields):
        try:
            value = self
            for field in fields:
                value = getattr(value, field)

            return value
        except (AttributeError, KeyError):
            return None

    def load(self, data):
        try:
            items = data.iteritems()
        except AttributeError:
            try:
                items = data._asdict().iteritems()
            except AttributeError:
                return

        for field, value in items:
            if isinstance(value, dict) or hasattr(value, "_asdict"):
                self.dict[field] = ModelObject(field).load(value)
            elif isinstance(value, (list, tuple, set)):
                items = []
                for item in value:
                    items.append(ModelObject(field).load(item))

                self.dict[field] = items
            else:
                self.dict[field] = value

        return self

    def to_dict(self):
        result = self.dict.copy()
        for field, value in result.iteritems():
            if isinstance(value, ModelObject):
                result[field] = value.to_dict()
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, ModelObject):
                        value[i] = item.to_dict()
        return result

    def json(self):
        return json.dumps(self.to_dict())


class ModelErrors(object):
    def __init__(self):
        self.pointers = {}
        self.exceptions = {}

    def __repr__(self):
        return str(self.exceptions)

    def json(self):
        return json.dumps(self.exceptions)

    def add(self, obj, field, ex, idx):
        name = obj.prefix + field
        self.exceptions[name] = {"index": idx, "exception": ex}
        try:
            field_pointer = self.pointers[obj][field]
        except KeyError:
            self.pointers[obj] = {idx: {name: ex}}
        else:
            field_pointer[idx] = {name: ex}


class Model(ModelObject):
    def __init__(self):
        super(Model, self).__init__()

    def report_error(self, obj, field, ex, idx=0):
        if not self.errors:
            self.errors = ModelErrors()

        self.errors.add(obj, field, ex, idx)

    def is_valid_item(self, obj, name, idx):
        try:
            return not name in self.errors.pointers[obj][idx]
        except (KeyError, AttributeError):
            return True

    @property
    def is_valid(self):
        try:
            return not self.errors.exceptions
        except AttributeError:
            return True


class DescribedItem(object):
    def __init__(self, modifier, validators, default, required):
        self.modifier = modifier
        self.validators = validators
        self.default = default
        self.required = required


class Field(DescribedItem):
    def __init__(self, field, modifier=None, validators=None, default=validation.NO_DEFAULT, required=False):
        super(Field, self).__init__(modifier, validators, default, required)
        self.field = field


class ObjectField(DescribedItem):
    def __init__(self, obj, field, modifier=None, validators=None, default=validation.NO_DEFAULT, required=False):
        super(ObjectField, self).__init__(modifier, validators, default, required)
        self.object = obj
        self.field = field


class ArrayField(DescribedItem):
    def __init__(self, array, field, modifier=None, validators=None, default=validation.NO_DEFAULT, required=False):
        super(ArrayField, self).__init__(modifier, validators, default, required)
        self.array = array
        self.field = field