# -*- coding: utf8 -*-

__author__ = "ardevelop"

import datetime
import re
import os
import copy
import uuid

from . import control, escape

EXTEND = "extend"
IGNORE = "ignore"
REQUIRE = "require"
END = "end"
IF = "if"
FOR = "for"
ELIF = "elif"
WHILE = "while"
ELSE = "else"
ESCAPED = "@"
ESCAPE = ":"
END_OF_STATEMENT = ":"
EMPTY = ""
ECHO = "echo"
CONTINUE = "continue"
BREAK = "break"
INCLUDE = "include"
SECTION = "section"
HTML = "html"
I18N = "\""
COMMENT_START = "<!--"
COMMENT_END = "-->"
WHITESPACES = " \t"
QUOTES = "\"'"
QUOTE_ESCAPE = "\\"
VALID_NAME = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_"
OBJECT_SEPARATOR = "."
COMMA = ","
NEWLINE = "\n"
CRLF = "\r\n"
BRACKETS_OPEN = "{[("
BRACKETS_CLOSE = "}])"

HEADER = [EXTEND, IGNORE, REQUIRE]
WITH_STATEMENT = [IF, ELIF, WHILE, FOR, ELSE]
MUST_EXIT = [IF, FOR, WHILE]
WITHOUT_ARGS = [CONTINUE, BREAK]

MARKUP_DETECTOR = re.compile("@(\"|([\(a-zA-Z_][0-9a-zA-Z_]*)|(<!--)|@|)")
COMPRESS = re.compile("[\r\n\t\s]+")


class ParseError(Exception):
    def __init__(self, filename, row, message):
        self.filename = filename
        self.row = row
        self.message = message

    def __str__(self):
        return "\n\tMessage  : %s\n\tTemplate : %s\n\tRow      : %s" % (
            self.message, self.filename or "<string>", self.row or "<no row information>")


class Command:
    def __init__(self, name, arg=None):
        self.name = name
        self.arg = arg

    def __repr__(self):
        return "(%s:%s)" % (self.name, self.arg)


class Section:
    def __init__(self, name=None):
        self.name = name
        self.commands = []

    def __iter__(self):
        return iter(self.commands)

    def __repr__(self):
        return "%s:%s" % (self.name, repr(self.commands))

    def __setitem__(self, key, value):
        self.commands[key] = value

    def append(self, name, arg=None):
        command = Command(name, arg)
        self.commands.append(command)
        return command

    def append_section(self, name):
        section = Section(name)
        self.commands.append(section)
        return section


class Template:
    @staticmethod
    def fromfile(filename):
        with open(filename) as fp:
            return Template(fp.read(), filename)

    def __init__(self, source, filename=None):
        self.ignore = False
        self.extend = None
        self.require = set()
        self.commands = Section()
        self._templates = None

        last = 0
        parse_header = True
        parsed_header = []
        end = len(source)
        branches = []
        sections = []
        branch = self.commands

        def get_row(idx):
            try:
                start = source.rindex(NEWLINE, 0, idx)
            except ValueError:
                start = 0
            try:
                stop = source.index(NEWLINE, idx)
            except ValueError:
                stop = end

            return source[start:stop].strip(CRLF + WHITESPACES)

        def compress(html):
            return COMPRESS.sub(" ", html)

        def get_statement(start):
            stop = start
            quote = None
            brackets = [0 for bracket in BRACKETS_OPEN]
            while True:
                try:
                    char = source[stop]
                    if char in QUOTES:
                        if not quote:
                            quote = char
                        elif char == quote:
                            quote = None
                    elif quote:
                        pass
                    elif char == END_OF_STATEMENT:
                        if not quote and sum(brackets) == 0:
                            stop += 1
                            return source[start:stop], stop
                    elif char in BRACKETS_OPEN:
                        brackets[BRACKETS_OPEN.index(char)] += 1
                    elif char in BRACKETS_CLOSE:
                        index = BRACKETS_CLOSE.index(char)
                        if brackets[index] > 0:
                            brackets[index] -= 1
                        else:
                            raise ParseError(filename, get_row(start), 'Close not open bracket "%s"' % char)

                    stop += 1
                except IndexError:
                    raise ParseError(filename, get_row(start), 'Cannot find end of statement')

        def get_echo(start):
            stop = start
            quote = None
            brackets = [0 for bracket in BRACKETS_OPEN]

            while True:
                try:
                    char = source[stop]
                    if char in QUOTES and sum(brackets) > 0:
                        if not quote:
                            quote = char
                        elif char == quote:
                            quote = None
                    elif quote or char in OBJECT_SEPARATOR:
                        pass
                    elif char in BRACKETS_OPEN:
                        brackets[BRACKETS_OPEN.index(char)] += 1
                    elif char in BRACKETS_CLOSE:
                        index = BRACKETS_CLOSE.index(char)
                        if brackets[index] > 0:
                            brackets[index] -= 1
                        else:
                            return source[start:stop], stop
                    elif char not in VALID_NAME:
                        if not quote and sum(brackets) == 0:
                            return source[start:stop], stop

                    stop += 1
                except IndexError:
                    return source[start:stop], stop

        def get_string(start):
            while True:
                try:
                    char = source[start]
                    if char in QUOTES:
                        start += 1
                        start_from = start
                        while True:
                            try:
                                stop = source.index(char, start_from)
                                if source[stop - 1] == QUOTE_ESCAPE:
                                    start_from = stop + 1
                                    continue

                                return source[start:stop], stop + 1
                            except ValueError:
                                raise ParseError(filename, get_row(start), "String end not found")
                    elif char in WHITESPACES:
                        start += 1
                    else:
                        raise IndexError()
                except IndexError:
                    raise ParseError(filename, get_row(start), "String start not found")

        def get_name(start):
            while True:
                try:
                    char = source[start]
                    if char in VALID_NAME:
                        stop = start
                        try:
                            while True:
                                if source[stop] not in VALID_NAME:
                                    return source[start:stop], stop

                                stop += 1
                        except IndexError:
                            return source[start:stop], stop
                    elif char in WHITESPACES:
                        start += 1
                    else:
                        raise IndexError()
                except IndexError:
                    raise ParseError(filename, get_row(start), "Name not found")

        def get_names(start):
            name, start = get_name(start)
            names = [name]
            while True:
                try:
                    char = source[start]
                    if char == COMMA:
                        name, start = get_name(start + 1)
                        names.append(name)
                    elif char in WHITESPACES:
                        start += 1
                    else:
                        raise IndexError()
                except (IndexError, ParseError):
                    return names, start

        for occurrence in MARKUP_DETECTOR.finditer(source):
            start, stop = occurrence.regs[1]

            if start >= last:
                keyword = source[start:stop]

                if keyword in HEADER:
                    if parse_header:
                        if keyword in parsed_header:
                            raise ParseError(filename, get_row(start),
                                             'Keyword "%s" must be used only one time in template' % keyword)

                        parsed_header.append(keyword)
                        if IGNORE == keyword:
                            self.ignore = True
                        elif EXTEND == keyword:
                            template, stop = get_string(stop + 1)
                            self.extend = template
                        else:  # REQUIRE == keyword
                            names, stop = get_names(stop + 1)
                            self.require |= set(names)

                        last = stop
                    else:
                        raise ParseError(filename, get_row(start),
                                         'Keyword "%s" must be in header of template' % keyword)

                else:
                    parse_header = False

                    if last < start:
                        html = compress(source[last:start - 1])
                        if html:
                            branch.append(HTML, html)

                    if EMPTY == keyword:
                        try:
                            exited = branches.pop()
                            if isinstance(exited, Section):
                                branch = sections.pop()
                            else:
                                branch.append(END)
                        except:
                            raise ParseError(filename, get_row(start), "Cannot exit statement from outside")
                    elif ESCAPED == keyword:
                        branch.append(HTML, ESCAPED)
                    elif INCLUDE == keyword:
                        template, stop = get_string(stop)
                        branch.append(INCLUDE, template)
                    elif SECTION == keyword:
                        name, stop = get_name(stop)
                        section = branch.append_section(name)
                        if source[stop] == END_OF_STATEMENT:
                            stop += 1
                            sections.append(branch)
                            branch = section
                            branches.append(section)
                    elif I18N == keyword:
                        text, stop = get_string(start)
                        branch.append(I18N, text)
                    elif COMMENT_START == keyword:
                        try:
                            stop = source.index(COMMENT_END, stop) + len(COMMENT_END)
                        except ValueError:
                            stop = end

                        last = stop
                        continue
                    elif keyword in WITHOUT_ARGS:
                        branch.append(keyword)
                    elif keyword in WITH_STATEMENT:
                        statement, stop = get_statement(start)
                        if keyword == ELSE and statement != keyword + END_OF_STATEMENT:
                            raise ParseError(filename, get_row(start),
                                             'Keyword "else" must by followed only with end of statement char.')
                        command = branch.append(keyword, statement)
                        if keyword in MUST_EXIT:
                            branches.append(command)
                    else:
                        code, stop = get_echo(start)
                        branch.append(ECHO, code)

                    last = stop

        if branches:
            not_exited = branches[0]
            if isinstance(not_exited, Section):
                raise ParseError(filename, None, 'Cannot find end of section "%s".' % not_exited.name)
            else:
                raise ParseError(filename, None, 'Cannot find end of statement "%s".' % not_exited.arg)

        if last < end:
            html = compress(source[last:end])
            if html:
                branch.append(HTML, html)

    def __iter__(self):
        return iter(self.commands)

    def __setitem__(self, key, value):
        self.commands[key] = value

    def build(self, templates=None):
        if self.extend:
            try:
                parent = templates[self.extend]
            except KeyError:
                raise Exception('Cannot extend template. Template "%s" not found.' % self.extend)

            sections = [command for command in self.commands if isinstance(command, Section)]
            extended = parent.build(templates)
            extended.require |= self.require

            for i, command in enumerate(extended):
                if isinstance(command, Section):
                    try:
                        extended[i] = [section for section in sections if section.name == command.name][0]
                    except IndexError:
                        pass
        else:
            extended = copy.deepcopy(self)

        extended._templates = templates or []
        return extended

    def compile(self, namespace, i18n=None):
        if self._templates is None:
            raise Exception("Template must be built before compile.")

        code = ""
        iterators = [iter(self)]
        indent = 1
        first = True
        echo_stack = []
        while True:
            try:
                command = iterators[-1].next()
            except StopIteration:
                iterators.pop()
            except IndexError:
                break
            else:
                if isinstance(command, Section):
                    iterators.append(iter(command))
                elif command.name == INCLUDE:
                    try:
                        template = self._templates[command.arg]
                    except KeyError:
                        raise Exception('Cannot include template. Template "%s" not found.' % command.arg)
                    else:
                        template = template.build(self._templates)
                        self.require |= template.require
                        iterators.append(iter(template))
                else:
                    if command.name in (ECHO, HTML, I18N):
                        if command.name == ECHO:
                            echo_stack.append("escape.utf8(" + command.arg + ")")
                        else:
                            if command.name == I18N:
                                command.arg = i18n.translate(command.arg) if i18n else command.arg

                            if echo_stack:
                                last = echo_stack[-1]
                                if last[0] == '"':
                                    echo_stack[-1] = last[:-1] + '%s"' % command.arg.replace('"', '\\"')
                                    continue

                            echo_stack.append('"%s"' % command.arg.replace('"', '\\"'))
                    else:
                        if echo_stack:
                            code += "%so%s=%s\n" % (indent * "\t", "" if first else "+", "+".join(echo_stack))
                            echo_stack = []

                        first = False

                        if command.name == END:
                            indent -= 1
                        elif command.name in (IF, WHILE, FOR):
                            code += "%s%s\n" % (indent * "\t", command.arg if command.arg else command + ":")
                            indent += 1
                        elif command.name in (ELIF, ELSE):
                            code += "%s%s\n" % ((indent - 1) * "\t", command.arg if command.arg else command + ":")
                        elif command.name in ("continue", "break"):
                            code += "%s%s\n" % (indent * "\t", command.name)

        returned = False

        if echo_stack:
            if first:
                code += "%sreturn %s\n" % (indent * "\t", "+".join(echo_stack))
                returned = True
            else:
                code += "%so+=%s\n" % (indent * "\t", "+".join(echo_stack))

        if not returned:
            code += "\treturn o"

        name = "_" + uuid.uuid4().hex
        definition = "def %s(app=None, request=None%s):\n" % (
            name, "".join([", %s=None" % require for require in self.require]))

        code = compile(definition + code, "<string>", "exec", dont_inherit=True)
        exec code in namespace

        return namespace[name]


class TemplatesLoader:
    def __init__(self, path, i18n=None, namespace=None):
        if i18n:
            i18n.select("ru")

        self.templates = {}
        self.namespace = {
            "control": control,
            "escape": escape,
            "datetime": datetime.datetime
        }
        if namespace:
            self.namespace.update(namespace)
        templates = {}
        path_size = len(path)

        def load(path):
            if os.path.isdir(path):
                for item in os.listdir(path):
                    load(os.path.join(path, item))
            else:
                if path.endswith(".html"):
                    key = path[path_size:].strip("/")
                    templates[key] = Template.fromfile(path)

        load(path)

        for key, template in templates.iteritems():
            if not template.ignore:
                self.templates[key] = template.build(templates).compile(self.namespace, i18n)

    def render(self, template_name, **kwargs):
        try:
            template = self.templates[template_name]
        except KeyError:
            raise Exception("Template not found.")

        return template(**kwargs)


__all__ = ["Template", "TemplatesLoader"]