from __future__ import absolute_import

__author__ = "ardevelop"

import urllib

import tornado.web
import tornado.httpserver
import tornado.ioloop

from . import i18n, template, validation, model as _model


class RequestArguments(object):
    __slots__ = "request", "container"

    def __init__(self, request):
        self.request = request
        self.container = {}

    def __getattr__(self, arg):
        try:
            return self.container[arg]
        except KeyError:
            self.container[arg] = value = self.request.get_argument(arg, None)
            return value

    #def __setattr__(self, arg, value):
    #    if arg not in ("request", "__container__"):
    #        self.__container__[arg] = value

    def __setitem__(self, arg, value):
        self.container[arg] = value

    def __getitem__(self, arg):
        try:
            return self.container[arg]
        except KeyError:
            self.container[arg] = value = self.request.get_argument(arg, None)
            return value

    def __repr__(self):
        return str(self.container)

    def get_array(self, arg):
        try:
            return self.container[arg]
        except KeyError:
            self.container[arg] = value = self.request.get_arguments(arg)
            return value

    def gen(self, **kwargs):
        args = {}
        for k, v in self.container.iteritems():
            if v is not None:
                args[k] = v

        for k, v in kwargs.iteritems():
            if v is not None:
                args[k] = v
            else:
                try:
                    del args[k]
                except KeyError:
                    pass

        return urllib.urlencode(args, True)


class RequestHandler(tornado.web.RequestHandler):
    def __init__(self, application, request, **kwargs):
        super(RequestHandler, self).__init__(application, request, **kwargs)

        self.args = RequestArguments(self)

    def render_string(self, template_name, **kwargs):
        return self.application.template.render(template_name, app=self.application, request=self, **kwargs)

    def route(self, name, value, modifier=None, validators=None,
              default=validation.NO_DEFAULT, required=False):
        valid, value, exception = self.application.validator.validate(value, modifier, validators, default, required)

        if not valid:
            raise tornado.web.HTTPError(404)

        self.args[name] = value

        return value

    def arg(self, name, modifier=None, validators=None, default=validation.NO_DEFAULT, required=False):
        valid, value, exception = self.application.validator.validate(self.get_argument(name, None),
                                                                      modifier, validators, default, required)

        if not valid:
            raise tornado.web.HTTPError(400)

        self.args[name] = value

        return value

    def model(self, constructor):
        """
        @rtype      : _model.Model
        @return     : the model from constructor
        """
        model = constructor()
        for item in model.describe():
            if isinstance(item, _model.Field):
                valid, val, ex = self.application.validator.validate(self.args[item.field], item.modifier,
                                                                     item.validators, item.default, item.required)
                model.set(item.field, val)
                if not valid:
                    model.report_error(model, item.field, ex)

            elif isinstance(item, _model.ObjectField):
                valid, val, ex = self.application.validator.validate(self.args[item.object + "." + item.field],
                                                                     item.modifier, item.validators, item.default,
                                                                     item.required)
                obj = model.get_or_create_object(item.object)
                obj.set(item.field, val)
                if not valid:
                    model.report_error(obj, item.field, ex)
            elif isinstance(item, _model.ArrayField):
                arr_items = self.args.get_array("%s.%s" % (item.array, item.field))
                for i, arr_item in enumerate(arr_items):
                    valid, val, ex = self.application.validator.validate(arr_item, item.modifier, item.validators,
                                                                         item.default, item.required)

                    obj = model.get_or_create_item(item.array, i)
                    obj.set(item.field, val)
                    if not valid:
                        model.report_error(obj, item.field, ex, i)

        model.resolve()

        return model

    @property
    def is_xhr(self):
        try:
            return self._is_xhr
        except AttributeError:
            try:
                self._is_xhr = "XMLHttpRequest" == self.request.headers["X-Requested-With"]
            except KeyError:
                self._is_xhr = False

            return self._is_xhr


class Application(tornado.web.Application):
    def __init__(self, default_host="", transforms=None, wsgi=False,
                 template_path=None, template_namespace=None, i18n_path=None, **settings):

        self.i18n = i18n.Dictionary(i18n_path) if i18n_path else None
        if template_path:
            self.template = template.TemplatesLoader(template_path, self.i18n, template_namespace)
        else:
            self.template = None
        self.validator = validation.Validator(self.i18n)

        super(Application, self).__init__(None, default_host, transforms, wsgi, **settings)

    def init(self, *args, **kwargs):
        pass

    def on_initialized(self):
        pass

    def run(self, port, address="0.0.0.0", workers=0, *args, **kwargs):
        self.port = port
        self.address = address

        server = tornado.httpserver.HTTPServer(self)
        server.bind(port, address)
        server.start(workers)
        self.init(*args, **kwargs)
        ioloop = tornado.ioloop.IOLoop.instance()
        ioloop.add_callback(self.on_initialized)
        ioloop.start()